import * as Discord from "discord.js";
const { prefix, token, api_key, debug_channel } = require("../config.json");
import * as ytdl from "ytdl-core";
import * as search from "youtube-search";
// eslint-disable-next-line no-unused-vars
import { Song, QueueConstruct } from "./model";

async function debug(message: Discord.Message | Discord.PartialMessage, something: string) {
  const channel = await client.channels.fetch(debug_channel) as Discord.TextChannel;
  channel.send(`[${new Date().toUTCString()}] Reply on ${message.content}:\n${something}`);
}

const opts: search.YouTubeSearchOptions = {
  maxResults: 1,
  key: api_key,
  type: "video"
};

const client = new Discord.Client();

const queue = new Map();

client.once("ready", () => {
  console.log("Ready!");
});

client.once("reconnecting", () => {
  console.log("Reconnecting!");
});

client.once("disconnect", () => {
  console.log("Disconnect!");
});

client.on("message", async message => {
  if (message.author.bot) return;
  if (!message.content.startsWith(prefix)) return;

  debug(message, `Received message: ${message.content}`);

  const serverQueue: QueueConstruct = queue.get(message.guild.id);

  if (message.content.startsWith(`${prefix}play`)) {

    execute(message, serverQueue);
    return;

  } else if (message.content.startsWith(`${prefix}skip`)) {

    skip(message, serverQueue);
    return;

  } else if (message.content.startsWith(`${prefix}stop`)) {

    stop(message, serverQueue);
    return;

  } else if (message.content.startsWith(`${prefix}queue`)) {

    let toBeSent = ">>> **Queue for breakn't-bot:**\n";
    serverQueue.songs.forEach((song, index) => {
      toBeSent += index === 0 ? `_Currently playing:_ ${song.title}\n` : `${index}. ${song.title}\n`;
    });
    message.channel.send(toBeSent);

    return;

  } else if (message.content.startsWith(`${prefix}help`)) {

    let toBeSent = `>>> **Help for breakn't-bot:**
    - \`}play <song name>\`: Searches for a song on youtube and plays it or adds it to the queue.
    - \`}skip\`: Skips the currently playing song.
    - \`}stop\`: Forces the breakn't-bot to leave the channel. Poor bot.
    - \`}queue\`: Shows the queue.
    - \`}remove <index>\`: Removes the song with the specified index from the queue.
    - \`}pause\`: Pause the playback.
    - \`}resume\`: Resume the playback.
    - \`}np\`: Shows info about the currently playing song.`;

    message.channel.send(toBeSent);

    return;

  } else if (message.content.startsWith(`${prefix}remove`)) {

    remove(message, serverQueue);

    return;

  } else if (message.content.startsWith(`${prefix}np`)) {

    const currentSong = serverQueue.songs[0];
    const songInfo = await ytdl.getInfo(currentSong.url);
    message.channel.send(`>>> _Currently playing:_ ${currentSong.title}
    ${(serverQueue.connection.dispatcher.streamTime/1000).toFixed(0)}/${songInfo.length_seconds}`);

    return;

  } else if (message.content.startsWith(`${prefix}pause`)) {

    serverQueue.connection.dispatcher.pause();

    return;

  } else if (message.content.startsWith(`${prefix}resume`)) {

    serverQueue.connection.dispatcher.resume();

    return;

  } else {

    message.channel.send("command validn't!");

  }
});

async function execute(message: Discord.Message | Discord.PartialMessage, serverQueue: QueueConstruct) {
  const [, ...args] = message.content.split(" ");

  debug(message, `Trying to find song: ${args.join(" ")}`);

  const voiceChannel = message.member.voice.channel;

  if (!voiceChannel)
    return message.channel.send(
      "You need to be in a voice channel to play music!"
    );

  const permissions = voiceChannel.permissionsFor(message.client.user);
  if (!permissions.has("CONNECT") || !permissions.has("SPEAK"))
    return message.channel.send(
      "I need the permissions to join and speak in your voice channel!"
    );
  
  search(args.join(" "), opts, async (err, results) => {
    if(err) return debug(message, "Error! " + err.name + ": " + err.message + ". Stacktrace: " + err.stack);
  
    const song: Song = {
      title: results[0].title,
      url: results[0].link
    };

    if (!serverQueue) {
      const queueContruct: QueueConstruct = {
        textChannel: message.channel,
        voiceChannel: voiceChannel,
        connection: null,
        songs: [],
        volume: 5,
        playing: true
      };

      queue.set(message.guild.id, queueContruct);

      queueContruct.songs.push(song);

      try {
        var connection = await voiceChannel.join();
        queueContruct.connection = connection;
        play(message.guild, queueContruct.songs[0], message);
      } catch (err) {
        debug(message, "Error! " + err.name + ": " + err.message + ". Stacktrace: " + err.stack);
        queue.delete(message.guild.id);
        return message.channel.send(err);
      }
    } else {
      debug(message, `Playing song ${song.title}`);
      serverQueue.songs.push(song);
      return message.channel.send(`**${song.title}** has been added to the queue!`);
    }
  });
}

function remove(message: Discord.Message | Discord.PartialMessage, serverQueue: QueueConstruct) {
  const [, ...args] = message.content.split(" ");
  if (/^\d+$/.test(args[0])) {
    const index = Number(args[0]);
    if (index < serverQueue.songs.length) {
      const deleted = serverQueue.songs.splice(index, 1)[0];
      message.channel.send(`Removed **${deleted.title}** from the queue`);
    } else {
      debug(message, "The entered value is outside the queue's range.");
    }
  } else {
    debug(message, "The entered value is not a valid number!");
  }
}

function skip(message: Discord.Message | Discord.PartialMessage, serverQueue: QueueConstruct) {
  if (!message.member.voice.channel) {
    debug(message, "No song to skip!");
    message.channel.send("You have to be in a voice channel to stop the music!");
    return;
  }
    
  if (!serverQueue) {
    message.channel.send("There is no song that I could skip!");
    debug(message, "No song to skip!");
    return;
  }
  serverQueue.connection.dispatcher.end();
}

function stop(message: Discord.Message | Discord.PartialMessage, serverQueue: QueueConstruct) {
  if (!message.member.voice.channel)
    return message.channel.send(
      "You have to be in a voice channel to stop the music!"
    );
  if (serverQueue.songs.length) {
    serverQueue.songs = [];
  }

  serverQueue.connection.dispatcher.end();
  serverQueue.voiceChannel.leave();
}

function play(guild: Discord.Guild, song: Song, message: Discord.Message | Discord.PartialMessage) {
  const serverQueue: QueueConstruct = queue.get(guild.id);
  if (!song) {
    serverQueue.voiceChannel.leave();
    queue.delete(guild.id);
    return;
  }

  let dispatcher: Discord.StreamDispatcher;

  function attemptPlay() {
    debug(message, "Attempting to play song!");

    dispatcher = serverQueue.connection
      .play(ytdl(song.url))
      .on("finish", () => {
        serverQueue.songs.shift();
        play(guild, serverQueue.songs[0], message);
      })
      .on("error", err => {
        debug(message, "Error playing song!");
        debug(message, "Error! " + err.name + ": " + err.message + ". Stacktrace: " + err.stack);
        attemptPlay();
      });

    debug(message, "Successfully avoided errors!");
    dispatcher.setVolumeLogarithmic(serverQueue.volume / 5);
    serverQueue.textChannel.send(`Started playing: **${song.title}**`);
  }

  attemptPlay();
  
}

client.login(token);